"""Check Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def is_float(string):
	"""
	Check if `string` can be converted
	into a float.

	Parameters
	----------
	string : str
		String to check.

	Returns
	-------
	check : bool
		True if `string` can be converted.
	"""
	try:
		float(string)
		return True
	except ValueError:
		return False

def is_int(string):
	"""
	Check if `string` can be converted
	into an int.

	Parameters
	----------
	string : str
		String to check.

	Returns
	-------
	check : bool
		True if `string` can be converted.
	"""
	try:
		int(string)
		return True
	except ValueError:
		return False

def check_XY(X=None, Y=None):
	"""
	Check if `X` and/or `Y` have the same length
	returning well-formatted ndarrays.
	Note this does not guarantee `X` and `Y`
	will work with any models as it does not rigorously
	inspect data types, only that it is in tabular format.

	Parameters
	----------
	X : array-like, default=None
		Array `X`.

	Y : array-like, default=None
		Array `Y`.

	Returns
	-------
	X : ndarray
		Returns validated `X`.
		If not provided, returns None.

	Y : ndarray
		Returns validated `Y`.
		If not provided, returns None.

	If both `X` and `Y` are provided or neither are
	provided, the results will be returned as a tuple
	in order: (X, Y).
	"""
	if X is not None and Y is not None:
		X, Y = check_X(X), check_Y(Y)
		if X.shape[0] != Y.shape[0]:
			raise ValueError("X and Y do not have the same length")
		return X, Y
	elif X is not None:
		return check_X(X), None
	elif Y is not None:
		return None, check_Y(Y)
	else:
		return None, None

def check_X(X):
	"""
	Convert `X` into an ndarray
	and allow conversion to a 2-dimensional array
	if currently 1-dimensional.

	Parameters
	----------
	X : array-like
		Array `X`.

	Returns
	-------
	X : ndarray
		Returns validated `X`.
	"""
	X = np.array(X)
	if X.shape[0] == X.size:
		X = X.reshape(-1,1)
	elif X.ndim < 2:
		raise ValueError("Array should be at least 2 dimensions")
	return X

def check_Y(Y):
	"""
	Convert `Y` into an ndarray.

	Parameters
	----------
	Y : array-like
		Array `Y`.

	Returns
	-------
	Y : ndarray
		Returns validated `Y`.
	"""
	Y = np.squeeze(np.array(Y))
	if Y.ndim == 0 and Y.size != 0:
		Y = Y.reshape(-1)
	if Y.shape[0] == Y.size:
		Y = Y.reshape(-1)
	if Y.ndim != 1 and Y.ndim != 2:
		raise ValueError("Array should be 1-dimensional or 2-dimensional")
	return Y



import os
from setuptools import setup, find_packages

def read(*paths):
	"""
	Build a file path from *paths* and return the contents.
	"""
	with open(os.path.join(*paths), 'r') as f:
		return f.read()

setup(
	name='sleepens',
	version='1.0.1',
	description='Sleep Classification using Ensemble Classification',
	long_description=(read('README.md') + '\n\n'),
	long_description_content_type="text/markdown",
	url='http://github.com/paradoxysm/sleepens',
	download_url = 'https://github.com/paradoxysm/sleepens/archive/1.0.1.tar.gz',
	author='paradoxysm',
	author_email='paradoxysm.dev@gmail.com',
	license='BSD-3-Clause',
	packages=find_packages(exclude=['tests']),
	include_package_data=True,
	install_requires=[
		'numpy',
		'scikit-learn',
		'joblib',
		'tqdm',
		'xlrd',
		'xlwt',
		'sonpy',
		'scipy',
	],
	python_requires='>=3.7, <4',
	classifiers=[
		'Development Status :: 4 - Beta',
		'Intended Audience :: Developers',
		'Natural Language :: English',
		'License :: OSI Approved :: BSD License',
		'Operating System :: OS Independent',
		'Programming Language :: Python',
		'Programming Language :: Python :: 3',
		'Programming Language :: Python :: 3.7',
		'Programming Language :: Python :: 3.8',
		'Programming Language :: Python :: 3.9',
		'Topic :: Software Development :: Libraries :: Python Modules',
		'Topic :: Scientific/Engineering :: Artificial Intelligence',
		'Topic :: Scientific/Engineering :: Information Analysis',
		'Intended Audience :: Science/Research',
	],
	keywords=['python', 'ml', 'ensemble', 'sleep', 'classification'],
	zip_safe=True)



"""Dataset"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import os
import numpy as np

from sleepens.utils import is_float, is_int

class Dataset:
	"""
	Dataset.

	Parameters
	----------
	name : string, default="noname"
		Name of the dataset. When the dataset is exported,
		this is the name of the file.

	features : array-like, shape=(n_features), default=[]
		The names of the features of the data.
		Length of `features` must match the number of columns
		in `data`.

	label_names : array_like, shape=(n_labels), default=[]
		The names of the labels of the data.
		Length of `label_names` must match the number of columns
		in `labels`.

	data : array_like, shape=(n_samples, n_features), default=[]
		The data for this Dataset.

	labels : array_like, shape=(n_samples, n_labels) default=[]
		The labels for this Dataset.
	"""
	def __init__(self, name="noname", features=[], label_names=[],
					data=[], labels=[]):
		self.name = name
		self.features = np.array(features)
		self.label_names = np.array(label_names)
		self.data = np.array(data)
		self.labels = np.array(labels)

	def split(self, name="noname", rows=None, cols=None, label=None):
		"""
		Create a new Dataset with a subset of the current Dataset.

		Parameters
		----------
		name : string, default="noname"
			The name of the new Dataset.

		rows : int, list of indices, None, default=None
			Rows of data to use. If None, take all rows.

		cols : int, list of indices, None, default=None
			Columns of data to use. If None, take all columns.

		label : int, list of indices, None, default=None
			Labels to use. If None, take all labels.

		Returns
		-------
		ds : Dataset
			The newly created Dataset.
		"""
		data = self.data if rows is None else self.data[rows]
		data = self.data if cols is None else data[:,cols]
		features = self.features if cols is None else self.features[cols]
		labels = self.labels if rows is None else self.labels[rows]
		labels = self.labels if label is None else labels[:,label]
		label_names = self.label_names if label is None else self.label_names[label]
		return Dataset(name, features, label_names, data, labels)

	def append(self, ds):
		"""
		Append a Dataset to this Dataset.
		The features and label names of the appending Dataset
		must match this Dataset. The number of columns for
		data and labels must also match.

		Parameters
		----------
		ds : Dataset
			The Dataset to append.

		Returns
		-------
		self : Dataset
			Returns this Dataset.
		"""
		features = ds.features
		if np.any(features != self.features):
			raise ValueError("Features do not match. Looking for " + \
								self.features + ". Got " + features + ".")
		data = ds.data
		if data.shape[1] != self.data.shape[1]:
			raise ValueError("Data columns do not match. Looking for " + \
								self.data.shape[1] + " columns. Got " + \
								data.shape[1] + ".")
		label_names = ds.label_names
		if np.any(label_names != self.label_names):
			raise ValueError("Label names do not match. Looking for " + \
								self.label_names + ". Got " + label_names + ".")
		labels = ds.labels
		if labels.shape[1] != self.labels.shape[1]:
			raise ValueError("Label columns do not match. Looking for " + \
								self.labels.shape[1] + " columns. Got " + \
								labels.shape[1] + ".")
		self.data = np.concatenate((self.data, data))
		self.labels = np.concatenate((self.labels, labels))
		return self

	def concatenate(self, ds, overlap=True):
		"""
		Concatenate a Dataset to this Dataset.
		The number of samples in both datasets must match.

		Parameters
		----------
		ds : Dataset
			The Dataset to append.

		overlap : bool, default=True
			Determine how to handle duplicate features or labels.
			When True, only use the features/labels copy in this
			Dataset. When False, use both.

		Returns
		-------
		self : Dataset
			Returns this Dataset.
		"""
		if len(self.data) == 0:
			self.features = ds.features
			self.data = ds.data
		elif len(self.data) == len(ds.data):
			if overlap : feature_overlap = np.isin(ds.features, self.features, invert=True)
			else : feature_overlap = np.arange(ds.features.shape[0])
			self.features = np.concatenate((self.features, ds.features[feature_overlap]))
			self.data = np.concatenate((self.data, ds.data[:,feature_overlap]), axis=1)
		if len(self.labels) == 0:
			self.label_names = ds.label_names
			self.labels = ds.labels
		elif len(self.labels) == len(ds.labels):
			if overlap : label_overlap = np.isin(ds.label_names, self.label_names, invert=True)
			else : label_overlap = np.arange(ds.label_names.shape[0])
			self.labels = self.labels.reshape(-1, len(self.label_names))
			self.label_names = np.concatenate((self.label_names, ds.label_names[label_overlap]))
			self.labels = np.concatenate((self.labels, ds.labels.reshape(-1, len(ds.label_names))[:,label_overlap]), axis=1)
		else : raise ValueError("Label lengths do not match")
		return self

	def clean(self):
		"""
		Clean the dataset for mismatched data and label lengths.
		Truncates the longer of data or labels to match.

		Returns
		-------
		self : Dataset.
			Returns this Dataset.
		"""
		if len(self.label_names) > 0:
			length = min([len(self.data), len(self.labels)])
			self.data = self.data[:length]
			self.labels = self.labels[:length]
		else:
			self.labels = np.array([])
		return self



"""Miscellaneous Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

def aggregate(data, method, axis=0):
	"""
	Summarize data based on the `method`.

	Parameters
	----------
	data : ndarray
		Data.

	method : {'mean', 'median', 'min', 'max', 'sum', 'count'}
		Method to summarize data.
		Must be one of:
		 - 'mean' : Mean of variances above threshold.
		 - 'median' : Median of variances above threshold.
		 - 'max' : Max of variances above threshold.
		 - 'min' : Min of variances above threshold.
		 - 'sum' : Sum of variances above threshold.
		 - 'count' : Count of variances above threshold.

	axis : int, default=0
		Axis to summarize along.

	Returns
	-------
	summarized : float
		The summarized data.
	"""
	if method == 'mean' : return np.mean(data, axis=axis)
	elif method == 'median' : return np.median(data, axis=axis)
	elif method == 'min' : return np.min(data, axis=axis)
	elif method == 'max' : return np.max(data, axis=axis)
	elif method == 'sum' : return np.sum(data, axis=axis)
	elif method == 'count': return data.shape[axis]
	else: raise ValueError("Method must be 'mean','median','min','max','sum','count'")

def determine_threshold(data, threshold, axis=0):
	"""
	Determine the real value threshold of the `data`
	given a specified `threshold` type.

	Parameters
	----------
	data : ndarray
		Data.

	threshold : {'mean', 'median', 'min', 'max', 'sum', 'count'}
		Method to determine threshold.
		Must be one of:
		 - 'mean' : Mean.
		 - 'median' : Median.
		 - '%<k>' : k-th percentile where <k> is an int.

	axis : int, default=0
		Axis to calculate along.

	Returns
	-------
	threshold : float or ndarray
		The threshold or list of thresholds.
	"""
	if isinstance(threshold, (float, int)) : pass
	elif threshold == 'mean': threshold = np.mean(data, axis=axis)
	elif threshold == 'median': threshold = np.median(data, axis=axis)
	elif threshold[0] == '%' and threshold[1:].isdigit():
		threshold = np.percentile(data, int(threshold[1:]), axis=axis)
	else : raise ValueError("Threshold must be 'mean','median','%<k>',float/int")
	return threshold



"""REMDrop Fix"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np

from sleepens.postprocess import default_map

def REMDrop(Y_hat, p, map=default_map, window=3, threshold=0.1):
	"""
	REM Drop Addon. Drop the rest of a REM episode
	if the probability declines past a moving average.

	Fix is conducted by determining if the probability of REM becomes
	lower than a moving average and dropping all subsequent REM states
	in favour of the next highest probable state.

	Parameters
	----------
	Y_hat : array-like, shape=(n_samples, n_classes)
		The raw prediction probabilities.

	p : array-like, shape=(n_samples,)
		The predictions to process. If no Addon
		processing was done prior to this, `p`
		corresponds with Y_hat.

	map : dict
		Mapping the label values to some
		set of integers.

	window : int, default=3
		Running average window size.

	threshold : float, default=0.1
		Maximum allowable drop in REM probability to
		maintain REM episode.

	Returns
	-------
	p : array-like, shape=(n_samples,)
		The post-processed predictions.
	"""
	for i in range(len(p)):
		if p[i] == map['R']:
			end = i
			while end < len(p) and p[end] == 3:
				end += 1
			if end - i > window:
				j, k = end-window, end
				post_avg, avg = 0, 0
				while k - i > window and avg - post_avg <= threshold:
					post_avg = np.mean(Y_hat[j:k,map['R']])
					j, k = j-1, k-1
					avg = np.mean(Y_hat[j:k,map['R']])
				if avg - post_avg > threshold:
					p[k:end] = np.argmax(Y_hat[k:end,:map['R']], axis=1)
			i = end
	return p



"""smrMAT I/O Interface"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from scipy.io import loadmat

from sleepens.io import DataObject, Dataset

name = "smrMAT"
standard = ".mat files exported by CED Spike2"
filetypes = [("MAT-files", "*.mat")]
type = "RAW"
tags = {'r'}

def read_data(filepath, channel):
	"""
	Read the data file at a specific data channel.

	Parameters
	----------
	filepath : path
		Path to the .mat file.

	channel : str
		Name of the channel in the .mat file.

	Returns
	-------
	dataobject : DataObject
		The DataObject containing the data from
		the specific channel.
	"""
	matfile = _load(filepath)
	fields = [f for f in matfile.keys() if '_Ch' in f]
	channels = [matfile[field][0][0][0][0] for field in fields]
	if channel in channels:
		field = fields[channels.index(channel)]
		try:
			data = matfile[field][0][0][8].flatten()
			resolution = matfile[field][0][0][2][0][0]
		except Exception:
			raise FileNotFoundError("An error occurred extracting from channel")
	else:
		raise FileNotFoundError("Channel named " + channel + " not found. Instead found: " + str(channels))
	return DataObject(name=channel, data=data, resolution=resolution)

def read_labels(filepath, channel, map={}):
	"""
	Read the data file at a specific label channel.

	Parameters
	----------
	filepath : path
		Path to the .mat file.

	channel : str
		Name of the channel in the .mat file.

	map : dict, default={}
		Mapping the label values to some
		set of integers.

	Returns
	-------
	dataobject : DataObject
		The DataObject containing the labels from
		the specific channel.
	"""
	matfile = _load(filepath)
	fields = [f for f in matfile.keys() if '_Ch' in f]
	channels = [matfile[field][0][0][0][0] for field in fields]
	if channel in channels:
		field = fields[channels.index(channel)]
		try:
			labels = matfile[field][0][0][7].flatten()[:-1]
			resolution = matfile[field][0][0][2][0][0]
			for k, v in map.items():
				labels[labels == k] = v
			labels = labels.astype(int)
		except Exception:
			raise FileNotFoundError("An error occurred extracting from channel")
	else:
		raise FileNotFoundError("Channel named " + channel + " not found. Instead found: " + str(channels))
	return DataObject(name=channel, data=labels, resolution=resolution)

def write(filepath, dataobjects):
	"""
	Write the dataset to a file.

	Parameters
	----------
	filepath : path
		Path to the .mat file to write.

	dataobjects : array-like of DataObject, shape=(n_channels,)
		DataObjects to write to the file. DataObjects
		with resolution set to -1 are assumed as labels.
	"""
	raise NotImplementedError("smrMAT cannot write to files")

def _load(filepath):
	"""
	Attempt to load the .mat file.

	Parameters
	----------
	filepath : path
		Path to the .mat file.

	Returns
	-------
	matfile : dict
		Dictionary with variable names as keys
		and matrices as values.
	"""
	try:
		matfile = loadmat(filepath)
	except:
		raise FileNotFoundError("No such file or directory: " + filepath)
	return matfile



import numpy as np

from sleepens.process import primary as prim
from sleepens.process import secondary as sec
from sleepens.utils.data import sampling as s
from sleepens.utils.data import normalize as n
from sleepens.io import Dataset

def process(data, labels, name, params, verbose=0):
	eeg, neck, mass = data
	if verbose > 0 : print("Extracting Features")
	if verbose > 1 : print("Detrending Data")
	eeg, neck, mass = _detrend(eeg, neck, mass)
	if verbose > 1 : print("Conducting FFT Analysis")
	eeg_fft = _fft(params['EEG_FFT'], params['EPOCH_SIZE'], eeg)[0]
	neck_fft, mass_fft = _fft(params['EMG_FFT'], params['EPOCH_SIZE'], neck, mass)
	if verbose > 1 : print("Calculating RMS")
	eeg_rms, neck_rms, mass_rms = _rms(params['EPOCH_SIZE'], eeg, neck, mass)
	if verbose > 1 : print("Calculating Percentile Mean")
	neck_prct, mass_prct = _prct(params['EPOCH_SIZE'], params['PRCTILE'], neck, mass)
	if verbose > 1 : print("Calculating Epoched Variance")
	neck_twitch, mass_twitch = _ep_var(params['EPOCHED_VAR'], params['EPOCH_SIZE'], neck, mass)
	if verbose > 1 : print("Calculating Spectral Entropy")
	eeg_entropy, neck_entropy, mass_entropy = _entropy(eeg_fft, neck_fft, mass_fft)
	if verbose > 1 : print("Calculating Frequency Bands")
	eeg_bands = sec.spectral_band(eeg_fft, bands=params['BANDS'], merge=params['BAND_MERGE'])
	if verbose > 1 : print("Calculating Frequency Ratios")
	ratios = sec.ratio(eeg_bands, ratios=params['RATIOS'])
	if verbose > 1 : print("Merging")
	emg_rms = _merge(mass_rms, neck_rms, feature='EMG RMS',
								method=params['EMG_RMS_MERGE'])
	emg_prct = _merge(mass_prct, neck_prct, feature='EMG PRCT',
								method=params['EMG_PRCT_MERGE'])
	emg_twitch = _merge(mass_twitch, neck_twitch, feature='EMG TWITCH',
								method=params['EMG_TWITCH_MERGE'])
	emg_entropy = _merge(mass_entropy, neck_entropy, feature='EMG ENTROPY',
								method=params['EMG_ENTROPY_MERGE'])
	features = [eeg_bands, ratios, eeg_rms, emg_rms, emg_prct,
					eeg_entropy, emg_entropy, emg_twitch]
	if verbose > 1 : print("Creating Dataset")
	ds = _create_ds(features, overlap=False)
	if verbose > 1 : print("Transforming Data")
	ds = sec.transform(ds, ops=params['TRANSFORM'])
	ds.data -= np.mean(ds.data, axis=0)
	if labels is not None:
		if verbose > 1 : print("Adding Labels")
		labels = Dataset(label_names=['LABELS'], labels=labels.data[:-1].reshape(-1,1))
		ds = ds.concatenate(labels)
	ds.name = name
	ds.clean()
	return ds

def _detrend(*channels):
	for c in channels:
		c.process(n.demean)
	return channels

def _fft(params, epoch_size, *channels):
	result = []
	for c in channels:
		fft = prim.fft(c, epoch_size,
						nperseg_factor=params['NPERSEG'],
						noverlap_factor=params['NOVERLAP'],
						detrend=params['DETREND'])
		result.append(fft)
	return result

def _rms(epoch_size, *channels):
	result = []
	for c in channels:
		rms = prim.rms(c, epoch_size)
		result.append(rms)
	return result

def _prct(epoch_size, k, *channels):
	result = []
	for c in channels:
		prct = prim.percentile_mean(c, epoch_size,
						k=k)
		result.append(prct)
	return result

def _ep_var(params, epoch_size, *channels):
	sub_epoch_size = params['SUBEPOCH_SIZE']
	threshold = params['THRESHOLD']
	merge = params['MERGE']
	result = []
	for c in channels:
		epvar = prim.epoched_variance(c, epoch_size,
						sub_epoch_size= sub_epoch_size,
						threshold=threshold, merge=merge)
		result.append(epvar)
	return result

def _entropy(*channels):
	result = []
	for c in channels:
		entropy = sec.spectral_entropy(c)
		result.append(entropy)
	return result

def _merge(channel1, channel2, feature='MERGE', method='mean'):
	if method is None:
		raise ValueError("Feature")
	merge = channel1.concatenate(channel2, overlap=False)
	return sec.merge(merge, feature=feature, method=method)

def _create_ds(features, overlap=False):
	if len(features) == 0 : return Dataset()
	ds = features.pop(0)
	for f in features:
		ds = ds.concatenate(f, overlap=overlap)
	return ds



"""Transformation Functions"""

# Authors: Jeffrey Wang
# License: BSD 3 clause

import numpy as np
from scipy import stats

def log_transform(data, base=np.e, axis=0):
	"""
	Transform the data into a logarithm.

	Parameters
	----------
	data : ndarray
		Data to be log transformed.

	base : {scalar, 'mean', 'median'}, default=np.e
		Log base for normalization. Must be one of:
		 - scalar : Constant base.
		 - 'mean' : Set the base to the mean of the data.
		 - 'median' : Set the base to the median of the data.

	axis : int, default=0
		Axis to transform along.

	Returns
	-------
	transform : ndarray
		Log transformed data.
	"""
	if base == 'mean':
		base = np.mean(data, axis=axis)
	elif base == 'median':
		base = np.median(data, axis=axis)
	return np.log(data) / np.log(base)

def boxcox_transform(data, lmbda=None):
	"""
	Transform the data by a Box-Cox power transformation.

	Parameters
	----------
	data : ndarray
		Data to be transformed.

	lmbda : scalar, default=np.e
		If lmbda is not None, do the transformation for that value.
		If lmbda is None, find the lambda that maximizes
		the log-likelihood function and return it as the
		second output argument.

	Returns
	-------
	transform : ndarray
		Transformed data.

	maxlog : float, optional
		If the lmbda parameter is None, maxlog is
		the lambda that maximizes the log-likelihood function.
	"""
	return stats.boxcox(data, lmbda=lmbda)

def power_transform(data, a):
	"""
	Transform the data into by a power.

	Parameters
	----------
	data : ndarray
		Data to be power transformed.

	a : scalar
		Exponent to use for transformation.

	Returns
	-------
	transform : ndarray
		Power transformed data.
	"""
	return np.sign(data) * np.power(np.absolute(data), a)
